# -*- coding: utf-8 -*-
from datetime import datetime, timedelta, date
import re

from bs4 import BeautifulSoup
from django.conf import settings
from django.utils import timezone
import requests
import simplejson as json


def isalambda(v):
    return isinstance(v, type(lambda: None)) and v.__name__ == '<lambda>'


class VkontakteParseError(Exception):
    pass


class VkontakteParser(object):

    content = ''

    def __init__(self, content=''):
        self.content = content

    @property
    def html(self):
        content = self.content
        # fix parsing html for audio tags in http://vk.com/wall-16297716_87985, http://vk.com/wall-16297716_182282?reply=182342
        # <!-- ->-> <!-- -<>-> <!-- -->
        content = re.sub(r'<!--.+?(?:->)?->', '', content)

        parts = content.split('<!>')

        for part in parts[5:]:
            if part[:4] == '<div':
                content = part
                break

        # concatenate preload container
        for part in parts[6:]:
            if part[:7] == '<!json>' and 'preload' in part:
                data = json.loads(part.replace('<!json>', ''))
                if not isinstance(data['preload'], bool):
                    content += data['preload'][0]
                break

        return content

    @property
    def content_bs(self):
        return BeautifulSoup(self.html)

    def request(self, *args, **kwargs):
        kwargs['headers'] = {'Accept-Language': 'ru-RU,ru;q=0.8'}

        args = list(args)
        if 'http' not in args[0]:
            args[0] = 'http://vk.com' + args[0]

        if 'method' in kwargs and kwargs.pop('method') == 'get':
            response = requests.get(*args, **kwargs)
        else:
            response = requests.post(*args, **kwargs)

        self.content = response.content.decode('windows-1251')
        return self

    def parse_time(self, text):
        return [int(v) for v in text.split(':')]

    def parse_date(self, date_text):
        date_words = date_text.split(' ')
        months = (u'', u'янв', u'фев', u'мар', u'апр', u'мая', u'июн', u'июл', u'авг', u'сен', u'окт', u'ноя', u'дек')
        hours = (u'', u'час', u'два', u'три', u'четыре', u'пять')
        minutes = (u'', u'минуту', u'две', u'три', u'четыре', u'пять')
        now = date.today()
        if u'сегодня в' in date_text:
            h, m = self.parse_time(date_words[-1])
            return datetime(now.year, now.month, now.day, h, m)
        elif u'вчера в' in date_text:
            h, m = self.parse_time(date_words[-1])
            return datetime(now.year, now.month, now.day, h, m) - timedelta(days=1)
        elif u'назад' == date_words[-1]:
            try:
                value = int(date_words[0])
            except:
                value = 0
            if date_words[-2].startswith(u'час'):
                # три часа назад
                # час назад
                return now - timedelta(hours=value or hours.index(date_words[0]))
            elif date_words[-2].startswith(u'минут'):
                # 4 минуты назад
                # минуту назад
                return now - timedelta(minutes=value or minutes.index(date_words[0]))
            elif date_words[-2].startswith(u'секунд'):
                # 10 секунд назад
                return now - timedelta(minutes=value)
        elif u'только что' == date_text:
            return now
        elif len(date_words) > 1:
            month = date_words[1][:3]
            if len(date_words) == 4:
                # 15 мая в 10:12
                h, m = self.parse_time(date_words[-1])
                value = datetime(now.year, months.index(month), int(date_words[0]), h, m)
                return value if value < now else datetime(now.year - 1, months.index(month), int(date_words[0]), h, m)

            elif len(date_words) == 3:
                # 31 дек 2011
                return datetime(int(date_words[2]), months.index(month), int(date_words[0]))

    def parse_container_likes(self, container, classname):
        try:
            value = container.find('span', {'class': classname}).text
            return value and int(value) or 0
        except Exception, e:
            raise VkontakteParseError("Error while parsing post likes value: %s" % e)

    def add_users(self, users, user_link, user_photo, user_add):
        if 'vkontakte_users' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured("Application 'vkontakte_users' not in INSTALLED_APPS")

        from vkontakte_users.models import User

        items = self.content_bs.findAll(*users)
        for item in items:
            user_link_container = user_link(item) if isalambda(user_link) else item.find(*user_link)
            user_photo_container = user_photo(item) if isalambda(user_photo) else item.find(*user_photo)

            user = User.remote.get_by_slug(user_link_container['href'][1:])
            if user:
                user.set_name(user_link_container.text)
                user.photo = user_photo_container['src']
                user.save()
                if isalambda(user_add):
                    user_add(user)
                else:
                    raise ValueError("Argument 'user_add' should be a lambda function, not %s" % user_add)

        return items



# -*- coding: utf-8 -*-
import logging

from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from m2m_history.fields import ManyToManyHistoryField
from vkontakte_users.models import User

from . import fields
from .models import VkontakteManager, VkontakteTimelineManager

log = logging.getLogger('vkontakte_api')


def get_or_create_group_or_user(remote_id):
    if remote_id > 0:
        Model = ContentType.objects.get(app_label='vkontakte_users', model='user').model_class()
    elif remote_id < 0:
        Model = ContentType.objects.get(app_label='vkontakte_groups', model='group').model_class()
    else:
        raise ValueError("remote_id shouldn't be equal to 0")

    return Model.objects.get_or_create(remote_id=abs(remote_id))[0]


class CountOffsetManagerMixin(VkontakteManager):

    def fetch(self, count=100, offset=0, **kwargs):
        count = int(count)
        if count > 100:
            raise ValueError("Attribute 'count' can not be more than 100")

        # count количество элементов, которое необходимо получить.
        if count:
            kwargs['count'] = count

        # offset смещение, необходимое для выборки определенного подмножества. По умолчанию — 0.
        # положительное число
        offset = int(offset)
        if offset:
            kwargs['offset'] = offset

        return super(CountOffsetManagerMixin, self).fetch(**kwargs)


class AfterBeforeManagerMixin(VkontakteTimelineManager):

    def fetch(self, before=None, after=None, **kwargs):
        if before and not after:
            raise ValueError("Attribute `before` should be specified with attribute `after`")
        if before and before < after:
            raise ValueError("Attribute `before` should be later, than attribute `after`")

        # special parameters
        if after:
            kwargs['after'] = after
        if before:
            kwargs['before'] = before

        return super(AfterBeforeManagerMixin, self).fetch(**kwargs)


class AuthorableModelMixin(models.Model):

    author_content_type = models.ForeignKey(
        ContentType, null=True, related_name='content_type_authors_%(app_label)s_%(class)ss')
    author_id = models.BigIntegerField(null=True, db_index=True)
    author = generic.GenericForeignKey('author_content_type', 'author_id')

    class Meta:
        abstract = True

    @property
    def by_group(self):
        return self.author_content_type.model == 'group' and self.author_content_type.app_label == 'vkontakte_groups'

    @property
    def by_user(self):
        return self.author_content_type.model == 'user' and self.author_content_type.app_label == 'vkontakte_users'

    def parse(self, response):
        if 'from_id' in response:
            self.author = get_or_create_group_or_user(response.pop('from_id'))
        super(AuthorableModelMixin, self).parse(response)


class OwnerableModelMixin(models.Model):

    owner_content_type = models.ForeignKey(
        ContentType, null=True, related_name='content_type_owners_%(app_label)s_%(class)ss')
    owner_id = models.BigIntegerField(null=True, db_index=True)
    owner = generic.GenericForeignKey('owner_content_type', 'owner_id')

    class Meta:
        abstract = True

    @property
    def on_group_wall(self):
        from vkontakte_groups.models import Group
        return isinstance(self.owner, Group)

    @property
    def on_user_wall(self):
        from vkontakte_users.models import User
        return isinstance(self.owner, User)

    @property
    def owner_remote_id(self):
        return self.get_owner_remote_id(self.owner)

    @classmethod
    def get_owner_remote_id(cls, owner):
        from vkontakte_users.models import User
        from vkontakte_groups.models import Group

        if isinstance(owner, User):
            return owner.remote_id
        elif isinstance(owner, Group):
            return -1 * owner.remote_id
        else:
            raise ValueError("Field owner should store User of Group, not %s" % owner.__class__)

    def parse(self, response):
        if 'owner_id' in response:
            self.owner = get_or_create_group_or_user(response.pop('owner_id'))
        super(OwnerableModelMixin, self).parse(response)


class LikableModelMixin(models.Model):

    likes_users = ManyToManyHistoryField(User, related_name='like_%(class)ss')
    likes_count = models.PositiveIntegerField(u'Likes', null=True, db_index=True)

    class Meta:
        abstract = True

    @property
    def likes_remote_type(self):
        raise NotImplementedError()

    @transaction.commit_on_success
    def fetch_likes(self, *args, **kwargs):

        kwargs['likes_type'] = self.likes_remote_type
        kwargs['item_id'] = self.remote_id_short
        kwargs['owner_id'] = self.owner_remote_id

        log.debug('Fetching likes of %s %s of owner "%s"' % (self._meta.module_name, self.remote_id, self.owner))

        ids = User.remote.fetch_likes_user_ids(*args, **kwargs)
        self.likes_users = User.remote.fetch(ids=ids, only_expired=True)

        # update self.likes_count
        likes_count = self.likes_users.count()
        if likes_count < self.likes_count:
            log.warning('Fetched ammount of like users less, than attribute `likes` of post "%s": %d < %d' % (
                self.remote_id, likes_count, self.likes_count))
        elif likes_count > self.likes_count:
            self.likes_count = likes_count
            self.save()

        return self.likes_users.all()

    def parse(self, response):
        if 'likes' in response:
            value = response.pop('likes')
            if isinstance(value, int):
                response['likes_count'] = value
            elif isinstance(value, dict) and 'count' in value:
                response['likes_count'] = value['count']
        super(LikableModelMixin, self).parse(response)


class RawModelMixin(models.Model):

    raw_json = fields.JSONField(default={}, null=True)

    class Meta:
        abstract = True

    def parse(self, response):
        self.raw_json = dict(response)
        super(RawModelMixin, self).parse(response)
